"""Reformats a paradigm table into a one word form per row csv file

© Kimmo Koskenniemi, 2017-2019

 This is free software under GPL 3 license.
"""

def step1():
    import csv, re, sys
    import twol.cfg as cfg

    version = cfg.timestamp(__file__)

    import argparse
    argparser = argparse.ArgumentParser(
        "python3 paratab2segcsv.py",
        description="Converts a tabular csv paradigm into"\
        " one example per row CSV file. Version {} ".format(version))
    argparser.add_argument(
        "input",
        default="ksk-paradigms.csv",
        help="Paradigm table as a CSV file")
    argparser.add_argument(
        "output",
        default="ksk-seg-examp.csv",
        help="One example per row paradigm as a CSV file")
    argparser.add_argument(
        "-s", "--morph-separator",
        default=".",
        help="Boundary between the morphs in a table cell")
    argparser.add_argument(
        "-d", "--csv-delimiter",
        default=",",
        help="CSV delimiter between the two fields, default is ','")
    argparser.add_argument(
        "-n", "--name-separator",
        default=".",
        help="Separator between morpheme names"\
        " in the morpheme list, default is '.'")
    argparser.add_argument(
        "-z", "--zero-symbol",
        default="Ø",
        help="Symbol to be inserted in word forms in order to"\
        " align them, default is Ø.  You are discouraged to change it.")
    args = argparser.parse_args()

    out_file = open(args.output, "w")
    writer = csv.DictWriter(out_file,
                            ["MORPHEMES","MORPHS"],
                            delimiter=args.csv_delimiter)
    writer.writeheader()
    d = {}

    morph_set = {}
    seg_ex_list = []
    with open(args.input, "r") as csvfile:
        reader = csv.DictReader(csvfile,
                                delimiter=args.csv_delimiter,
                                skipinitialspace=True)
        for row in reader:
            if row["ID"].startswith("?"):
                continue
            # process each cell of the row
            for column_label, words in row.items(): 
                if (not words) or (column_label in {"ID", "KSK"}) \
                   or ("STM" not in column_label):
                    continue
                morpheme_list = column_label.split(args.name_separator)
                if morpheme_list[0] == 'STM':
                    morpheme_list[0] = row['ID']
                words_clean = re.sub(r'[][()]', '', words)
                word_list = re.split(r"\s+", words_clean)
                for morphs in word_list:
                    if not morphs or morphs.find('*') >= 0:
                        continue
                    d["MORPHEMES"] = args.name_separator.join(morpheme_list).strip()
                    d["MORPHS"] = morphs
                    writer.writerow(d)
    out_file.close()
    return

if __name__ == "__main__":
    step1()
    



""" Forms morphophonemic representations out of zero-filled example word forms

    This program is free software under GPL 3 license

    Copyright Kimmo Koskenniemi 2017-2020
"""

def main():
    import re
    import csv
    import argparse
    import twol.cfg as cfg
    version = cfg.timestamp(__file__)

    argparser = argparse.ArgumentParser(
        "python3 raw2named.py",
        description="Renames raw morphophonemes. Version {} ".format(version))
    argparser.add_argument(
        "input",
        default="demo-raw.csv",
        help="aligned examples as a CSV file")
    argparser.add_argument(
        "output",
        default="demo-renamed.pstr",
        help="renamed examples as a space separated pair symbol strings")
    argparser.add_argument(
        "names",
        default="demo-renaming.csv",
        help="mapping from raw to neat morphophonemes as a CSV file,"\
        " default is ','")
    argparser.add_argument(
        "-d", "--delimiter",
        default=",",
        help="delimiter between raw name and new name fields, default is ','")
    argparser.add_argument(
        "-n", "--name-separator",
        default=".",
        help="Separator between morpheme names in the morpheme list,"\
        " default is '.'")
    argparser.add_argument(
        "-F", "--add-features",
        default=False, action="store_true",
        help="add affix morpheme names to the pairstring representation")
    argparser.add_argument(
        "-v", "--verbosity",
        default=0,
        type=int,
        help="level of diagnostic and debugging output")
    args = argparser.parse_args()

    import twol.cfg as cfg
    cfg.verbosity = args.verbosity

    mphon_name = { }

    # Read in the namefile is a CSV file which contains three fields:
    # 1. the raw (old) name for the mophophoneme
    # 2. a neat (new) name for the morphophoneme
    # 3. Comments documenting typical occurrences of the morphophoneme
    with open(args.names) as namefile:
        reader = csv.reader(namefile,
                            delimiter=args.delimiter,
                            skipinitialspace=True)
        for row in reader:
            if not row or (not row[0].strip()):
                continue
            if len(row) < 2:
                print("*** TOO FEW FIELDS IN:", row)
                continue
            if row[1].strip():
                mphon_name[row[0].strip()] = row[1].strip()

    #print(mphon_name)###

    outfil = open(args.output, "w")

    with open(args.input) as csvfile:
        reader = csv.DictReader(csvfile,
                                delimiter=args.delimiter,
                                skipinitialspace=True)
        for row in reader:
            zero_filled_str = row["ZEROFILLED"].strip().replace(".", "")
            raw_str = row["RAW"].strip()
            raw_lst = raw_str.split(" ")
            pairsym_lst = []
            if cfg.verbosity >= 20:
                print(row)
                print("raw_lst:", raw_lst)
            if len(raw_lst) != len(zero_filled_str):
                print("** LENGTHS DISAGREE **", raw_lst, zero_filled_str)
                continue

            for raw_insym, outsym in zip(raw_lst, zero_filled_str):
                if raw_insym == outsym:
                    psym = raw_insym
                else:
                    clean_insym = mphon_name.get(raw_insym, raw_insym)
                    psym = clean_insym + ":" + outsym
                pairsym_lst.append(psym)
            if args.add_features:
                morpheme_lst = row["MORPHEMES"].strip().split(args.name_separator)
                for morpheme in morpheme_lst[1:]:
                    pairsym_lst.append(morpheme + ":Ø")
            pairsym_str = " ".join(pairsym_lst)
            print(pairsym_str, file=outfil)

    return

if __name__ == "__main__":
    main()



"""alphabet.py

Processes an alphabet definition file so that it can be used both by
aligner.py and by multialign.py.  In particular, it computes weights
for phoneme pairs or morphophonemes for the purposes of weighted
alignment.

Copyright 2017-2020, Kimmo Koskenniemi

This is free software according to GNU GPL 3 license.
"""

import sys, re
import twol.cfg as cfg

cost_of_zero_c = 25
"""Additional weight for a subset of consonants if Ø belongs to it."""

cost_of_zero_v = 10
"""Additional weight for a subset of vowels if Ø belongs to it."""

feature_lst_lst = [["Zero"], ["Zero"], ["Zero"], ["Zero"], ["Zero"], ["Zero"]]
                             # six lists, one for each group of features

feature_bitpos = {}
"""The bit position of the feature within the 16 bit field of the group."""

mphon_to_binint_cache = {}
"""A 6*16 bit vector which represents the feature sets of this phoneme."""

binint_to_mphon_set = {}
"""For a binint, it gives the set of phoneme set previously stored for
a 96 bit integer."""

vowel_set = set()
"""The set of vowels including semivowels."""

consonant_set = set()
"""The set of consonants including semivowels."""

mphon_weight_cache = {}
"""A cache for morphophoneme weights."""

phoneme_set_weight_cache = {}
"""Given weights for phoneme sets represented by "".join(sorted(set(MPHON)))"""

for_definitions_lst = []
"""A list of ... FOR X IN ... definitions to be used in metric.py"""

exception_lst = []
"""A list of weighting exceptions to be used in metric.py"""

def spaced_bin_int(intg):
    """binint to human readable string conversion."""
    bs = "{:096b}".format(intg)
    spaced_str = (bs[0:16] + " " + bs[16:32] + " " + bs[32:48] +
                  " " + bs[48:64] + " " + bs[64:80] + " " + bs[80:96])
    return spaced_str

def mphon_to_binint(mphon):
    """Converts a morphophoneme into a binary integer that represents it.

    :param mphon: A string of phonemes and possibly Øs
    :type mphon: str
    :return: A binary integer which represents the morphophoneme
    :rtype: int
    """
    if cfg.verbosity >= 25:
        print("mphon_to_binint({})".format(mphon))
    if not mphon:
        return 0
    if mphon in mphon_to_binint_cache:
        return mphon_to_binint_cache[mphon]
    if len(mphon) == 1:
        msg = "** '{}' NOT IN ALPHABET\n\n".format(mphon)
        exit(msg)
    old = mphon[:-1]
    new = mphon[-1:]
    if new not in mphon_to_binint_cache:
        msg = "** '{}' IN '{}' NOT IN ALPHABET\n\n".format(new, mphon)
        exit(msg)
    if old in mphon_to_binint_cache:
        binint =  mphon_to_binint_cache[old] | mphon_to_binint_cache[new]
        mphon_to_binint_cache[mphon] = binint
        return binint
    else:
        return mphon_to_binint(old) | mphon_to_binint_cache[new]
    
def mphon_is_valid(mphon):
    """Tests whether a set of phonemes is possible.

    :param mphon: A sequence of phonemes a.k.a. morphophoneme, e.g. 'ij'.  
    :type mphon: str
    
    :return: True if the set consists either of vowels (and semivowels and Øs) \
    or of consonants (and semivowels and Øs), otherwise False.

    :rtype: boolean
    """

    binint = mphon_to_binint(mphon)
    if cfg.verbosity >= 30:
        print("mphon_is_valid({}) == {}".format(mphon, spaced_bin_int(binint)) )
    if not(~ binint & 0xffffffffffffffffffffffff): # (U,U,U,U,U,U)
        return False
    else:
        return True
    
def mphon_weight(mphon):
    """Returns the weight of a morphophoneme

    :param mphon: A sequence of phonemes a.k.a. morphophoneme, e.g. 'ij'
    :type mphon: str
    
    :return: The weight of mphon based on the phonological features of its members 
    :rtype: float
    """
    global weight_c1, weight_c2, weight_c3, weight_v1, weight_v2, weight_v3
    if re.fullmatch(r"[Ø]+", mphon):
        return cfg.all_zero_weight
    phon_set_str = "".join(sorted(set(mphon)))
    if phon_set_str in phoneme_set_weight_cache:
        return phoneme_set_weight_cache[phon_set_str]
    if mphon in mphon_weight_cache:
        return mphon_weight_cache[mphon]
    mphon_int = mphon_to_binint_cache[mphon]
    if cfg.verbosity >= 25:
        print("\nmphon_to_binint_cache[{}] = {}".format(mphon,
                                                      spaced_bin_int(mphon_int)))
    w_cons = 999999
    w_vow = 999999
    high = mphon_int >> 48               # extract the 48 cons feature bits
    low = mphon_int & 0xffffffffffff     # extract the 48 voc feature bits
    if cfg.verbosity >= 25:
        print("{:048b}".format(high), "high")
        print("{:048b}".format(low), "low")
    if high != 0xffffffffffff:
        c1 = high >> 32                  # place of articulation set
        c2 = (high >> 16) & 0xffff       # voicing set
        c3 = high & 0xffff               # manner of articulation set
        if cfg.verbosity >= 25:
            print("{:012b}, {:012b}, {:012b}".format(c1, c2, c3))
        w_cons = weight_c1[c1] + weight_c2[c2] + weight_c3[c3]
        if cfg.verbosity >= 25:
            print("\nmphon_weight info of a cons set:", hex(c1), weight_c1[c1],
                  hex(c2), weight_c2[c2], hex(c3), weight_c3[c3])
    if low != 0xffffffffffff:
        v1 = low >> 32                   # tongue height
        v2 = (low >> 16) & 0xffff        # backness
        v3 = low & 0xffff                # rounding
        if cfg.verbosity >= 25:
            print("{:012b}, {:012b}, {:012b}".format(v1, v2, v3))
        w_vow = weight_v1[v1] + weight_v2[v2] + weight_v3[v3]
        if cfg.verbosity >= 25:
            print("\nmphon_weight info of a vowel set:", hex(v1), weight_v1[v1],
                  hex(v2), weight_v2[v2], hex(v3), weight_v3[v3])
    w = min(w_cons, w_vow)
    if cfg.verbosity >= 25:
        print("\nmphon_int[{}]  = {}".format(mphon, spaced_bin_int(mphon_int)))
        print("mphon_weight[{}] = {}".format(mphon, w))
    mphon_weight_cache[mphon] = w
    phoneme_set_weight_cache[phon_set_str] = w
    return w

def read_alphabet(file_name):
    """Reads phoneme features, feature subsets with weights, and other definitions
    
    :param file_name: Name of the file that contains the alphabet definition
    :type file_name: str

    Stores the computed sets and weights for feature subsets and other
    info in module variables for the use of other modules.

    """
    global weight_c1, weight_c2, weight_c3, weight_v1, weight_v2, weight_v3

    subset_lst = []              # list of pairs (set, weight)
    feature_group = {}           # the group to which this feature belongs
    features_of_phoneme = {}     # tuples of six features for a phoneme
    with open(file_name, "r") as f:
        features_of_phoneme["Ø"] = ("Zero","Zero","Zero","Zero","Zero","Zero")
        i = 0
        for line_nl in f:
            i += 1
            line = line_nl.split("#")[0].strip()
            if not line:
                continue
            mat_phon_feat = re.fullmatch(
                r":?(?P<symbol>(\w|')):? *= *(?P<feats>\w*( *, *\w*)+)",
                line)
            if mat_phon_feat:
                # it defines features of a phoneme
                r_lst = [feat.strip() for feat in mat_phon_feat.group("feats").split(",")]
                if len(r_lst) != 6:
                    msg = "** WRONG NUMBER OF FEATURES ON LINE {}:\n{}"
                    sys.exit(msg.format(i, line))
                if mat_phon_feat.group("symbol") in features_of_phoneme:
                    msg = "** {} ALREADY DEFINED. LINE {}:\n{}"
                    sys.exit(msg.format(mat_phon_feat.group("symbol"), i, line_nl))
                features_of_phoneme[mat_phon_feat.group("symbol")] = tuple(r_lst)
                for ls, feat in zip(feature_lst_lst, r_lst):
                    if not feat in ls and feat:
                        ls.append(feat)
                continue
            mat_feat_set = re.fullmatch(
                r"(?P<elements>\w\w\w+( +\w\w\w+)+) *= *(?P<weight>[0-9]+)",
                line)
            if mat_feat_set:
                # it defines a subset of features and its weight
                l_lst = mat_feat_set.group("elements").split()
                subset_lst.append((set(l_lst), int(mat_feat_set.group("weight"))))
                continue
            mat_phon_set = re.fullmatch(
                r"(?P<weight>[0-9]+) *= *(?P<elements>(\w|')+( +(\w|')+)+)",
                line)
            if mat_phon_set:
                # it defines a subset of features and its weight
                phon_set_str_lst = mat_phon_set.group("elements").split()
                weight = int(mat_phon_set.group("weight"))
                for phon_str in phon_set_str_lst:
                    phon_set_str = "".join(sorted(set(phon_str)))
                    phoneme_set_weight_cache[phon_set_str] = weight
                continue
            mat_zero = re.fullmatch(
                r"Zero *[+]= *(?P<consw>[0-9]+) +(?P<voww>[0-9]+)",
                line)
            if mat_zero:
                # it defines the cost of including a Zero in feature sets
                cost_of_zero_c = int(mat_zero.group("consw"))
                cost_of_zero_v = int(mat_zero.group("voww"))
                continue
            mat_for_in = re.fullmatch(
                r"(?P<expr>(\w:\w +)*\w:\w::[0-9]+) +FOR +(?P<var>\w+) +IN +(?P<set>\w+)",
                line)
            if mat_for_in:
                # it defines a FOR IN definition
                for_definitions_lst.append((mat_for_in.group("expr"),
                                            mat_for_in.group("var"),
                                            mat_for_in.group("set")))
                continue
            mat_exception = re.fullmatch("(?P<expr>(\w:\w +)*\w:\w::[0-9]+)", line)
            if mat_exception:
                # it defines an exception list
                exception_lst.append(mat_exception.group("expr"))
                continue
            msg = "** INCORRECT ALPHABET DEFINITON LINE {}:\n {}"
            sys.exit(msg.format(i, line_nl))

    feature_set_lst = [set(lst) for lst in feature_lst_lst]
    #
    # now the alphabet data has been read in and extracted
    #
    if cfg.verbosity >= 20:
        print("\ncost_of_zero_c:", cost_of_zero_c)
        print("\ncost_of_zero_v:", cost_of_zero_v)
        print("\nfeature_set_lst:", feature_set_lst)
        print("\nfeature_lst_lst:", feature_lst_lst)
        print("\nsubset_lst:", subset_lst)
        print("\nfeatures_of_phoneme:", features_of_phoneme)
        print("\nfor_definitions_lst:", for_definitions_lst)
        print("\nexception_lst:", exception_lst)
    #
    # find the groups and bit positions of individual features
    #
    i = 0
    for feature_lst in feature_lst_lst:
        j = 0
        for feature in feature_lst:
            feature_group[feature] = i
            feature_bitpos[feature] = j
            j +=1
        i += 1
    del feature_group["Zero"]   # feature Zero belongs all groups and
                                # needs special care
    if cfg.verbosity >= 20:
        print("\nfeature_group:", feature_group)
        print("\nfeature_bitpos:", feature_bitpos)
    #
    # An integer for each phoneme.  The integer represents six sets
    # (with one element in each set).  The sets are 16 bit long fields
    # of the binary representation of the integer, total of 96 bits.
    # Each set has a bit position reserved for each feature in the
    # respective group.  These integers or bit vectors can be combined
    # with each other using bit operations.
    #
    for phoneme, features in features_of_phoneme.items():
        intset = 0
        for feature in features:
            if feature:
                bit_pos = feature_bitpos[feature]
                bin_set = 1 << bit_pos
                intset = (intset << 16) | bin_set
            else:
                intset = (intset << 16) | 0xffff
        mphon_to_binint_cache[phoneme] = intset
        if intset not in binint_to_mphon_set:
            binint_to_mphon_set[intset] = set()
        binint_to_mphon_set[intset].add(phoneme)
    if cfg.verbosity >= 20:
        lst = [fon + "=" + spaced_bin_int(intg) for fon, intg in
               sorted(mphon_to_binint_cache.items())]
        s = "\n".join(lst)
        print("\nmphon_to_binint_cache")
        print(s)
        t = "\n".join([spaced_bin_int(intg) + " = " + str(fon_set)
                       for intg, fon_set in sorted(binint_to_mphon_set.items())])
        print("\nbinint_to_mphon_set")
        print(t)
    #
    # sets of vowels and consonants
    #
    for phoneme, features in features_of_phoneme.items():
        if features[0] and features[1] and features[2]:
            consonant_set.add(phoneme)
        elif features[3] and features[4] and features[5]:
            vowel_set.add(phoneme)    
    #
    # convert the subsets into integers which represent bit vectors of the sets
    #
    subset_bin_lst = []
    for subset, weight in subset_lst:
        group_lst = list(set([feature_group[feature] for feature in subset]))
        if len(group_lst) == 1:
            group = group_lst[0]
        else:
            sys.exit("** FEATURES FROM SEVERAL GROUPS: {} = {}".format(subset, group_lst))
        #print("\ngroup:", subset, weight, group) ###
        bin_set = 0
        for feat in subset:
            bin_set = bin_set | (1 << feature_bitpos[feat])
        if cfg.verbosity > 25:
            print("\nsubset, bin_set, weight, group:", subset, bin(bin_set), weight, group)
        subset_bin_lst.append((bin_set, weight, group, bin(bin_set)))
    if cfg.verbosity > 20:
        print("\nsubset_bin_lst:", subset_bin_lst)
    #
    # compute weights for all possible feature sets in each of the six groups
    #
    i = 0
    weight_dict_lst = []
    for feature_lst in feature_lst_lst:
        weight_dict = {}
        weight_dict[0] =  -1 # for the empty set
        weight_dict[1] = -1   # {"Zero"}
        weight_dict[0xffff] = 999999 # for the universal set U
        l = len(feature_lst)
        for j in range(2, 1 << l, 2):
            w = 100
            for subset_bin, weight, group, bin_str in subset_bin_lst:
                if cfg.verbosity >= 25:
                    print("\nsubset_bin, weight, group, bin_str, i:",
                          bin(j), weight, group, bin_str, i)
                test = ~(subset_bin | ~j)
                if cfg.verbosity >= 25:
                    print(">>> test:", bin(test)) ###
                if group == i and not test and weight < w:
                    w = weight
            weight_dict[j] = w
            weight_dict[j+1] = w + (cost_of_zero_c if i < 3 else cost_of_zero_v)
        for j in range(1, l):
            weight_dict[1 << j] = 0
            weight_dict[(1 << j) +1] = (cost_of_zero_c if i < 3 else cost_of_zero_v)
        if cfg.verbosity > 20:
            print("\nweight_dict[{}]:".format(i), weight_dict)
            #for set_int, weight in weight_dict.items():
            #    for phoneme, feature_tuple in features_of_phoneme.items():
            #        feature = feature_tuple[i]
            #        ***
                
        weight_dict_lst.append(weight_dict)
        i += 1
    (weight_c1, weight_c2, weight_c3, weight_v1, weight_v2, weight_v3) = weight_dict_lst
    return

if __name__ == "__main__":
    cfg.verbosity = 1
    read_alphabet("alphabet-test.text")
    mphon_is_valid("ei")
    print(mphon_weight("ei"))



"""multialign.py

Aligns two words or more (or morphs) by adding some zero symbols so
that phonemes in corresponding positions are optimally similar.

Copyright 2020, Kimmo Koskenniemi

This is free software according to GNU GPL 3 license.

"""

import hfst
import grapheme
import twol.fs as fs
import twol.cfg as cfg
import twol.twbt as twbt
import twol.alphabet as alphabet


def init(alphabet_file_name, all_zero_weight=1):
    """Initializes multialign by initializing the alphabet module
alphabet_file_name -- an alphabet definition as described in https://pytwolc.readthedocs.io/en/latest/alignment.html#alphabet

all_zero_weight -- penalty for an intermediate morphophoneme of only zeros (which will get a non-zero component in the final morphophoneme)"""
    alphabet.read_alphabet(alphabet_file_name)
    cfg.all_zero_weight = all_zero_weight
    return


def print_result(aligned_result, comments, weights, layout="horizontal"):
    """Prints the result of the alignment in one of the three formats

aligned_result -- tuple of the weight and a list of aligned words where each aligned word is a list of 

comments -- possible comments which will be passed over

weights -- whether to print also the overall weight of this alignment

layout -- one of "horizontal" (a sequence of morphophonemes on a single line), "vertical" (each zero-filled word on a line of its own) or "list" (all zero-filled words on a single line)"""

    weight, aligned_words_lst = aligned_result
    if cfg.verbosity >= 10:
        print("aligned_result", aligned_result)
    
    if layout == "horizontal":
        lgth = grapheme.length(aligned_words_lst[0])
        mphon_lst = []
        for i in range(lgth):
            lst = []
            for aligned_word in aligned_words_lst:
                symbol = grapheme.slice(aligned_word, start=i, end=i+1)
                lst.append(symbol)
            if len(set(lst)) == 1:
                mphon_str = lst[0]  # abbreviate if all identical
            else:
                mphon_str = "".join(lst)
            mphon_lst.append(mphon_str)
        zstem_pairsym_str = " ".join(mphon_lst)

        mphonemic_str = " ".join(mphon_lst)
        if weights:
            print(mphonemic_str.ljust(40), weight)
        else:
            print(mphonemic_str)
    elif layout == "vertical":
        print("\n".join(aligned_words_lst))
        print()
    elif layout == "list":
        print(" ".join(aligned_words_lst))
    return


def accum_input_labels(fst, separator=""):
    """Encode, weight and prune a transducer

fst -- transducer to be processed, input labels are strings of alphabet symbols and output labels are single alphabet symbols

separator -- null string or a symbol not part of the alphabet

Returns a transducer where input labels of thrasitions are concatenations of the input label and the output label of the original transition, the weights are according to the weights of the resulting morphophonemes and all transitions with invalid morphophoneme labels are discarded.
"""
    if cfg.verbosity >= 10:
        print("to be accumulated:\n", fst)
    bfst = hfst.HfstBasicTransducer(fst)
    result_bfst = hfst.HfstBasicTransducer()
    for state in bfst.states():
        result_bfst.add_state(state)
        if bfst.is_final_state(state):
            weight = bfst.get_final_weight(state)
            result_bfst.set_final_weight(state, weight)
        for arc in bfst.transitions(state):
            tostate = arc.get_target_state()
            insym = arc.get_input_symbol()
            outsym = arc.get_output_symbol()
            weight = arc.get_weight()
            new_insym = insym + separator + outsym
            if cfg.verbosity >= 25:
                print("arc", state, tostate, insym, outsym, weight)
            if not alphabet.mphon_is_valid(new_insym):
                continue
            new_weight = alphabet.mphon_weight(new_insym)
            result_arc = hfst.HfstBasicTransition(tostate,
                                                  new_insym,
                                                  new_insym,
                                                  new_weight)
            result_bfst.add_transition(state, result_arc)
            if cfg.verbosity >= 25:
                print("after addition of transition:\n", result_bfst)
    result_fst = hfst.HfstTransducer(result_bfst)
    result_fst.minimize()
    if cfg.verbosity >= 10:
        print("accumulated fst:\n", result_fst)
    return result_fst


def list_of_aligned_words(mphon_lst):
    """Converts a list of morphophonemes into a list of aligned words

mphon_lst -- list of same length morphophonemes, e.g.
    ["lll", "ooo", "vvv", "ieØ"]

Returns a list of words constructed out of the 1st, 2nd ... alphabetic
symbols of the morphophonemes, e.g.  ["lll", "ooo", "vvv", "ieØ"] -->
["lovi", "love", "lovØ"]

    """
    if not mphon_lst:
        return []
    lgth = grapheme.length(mphon_lst[0])
    res = []
    for i in range(lgth):
        syms = [grapheme.slice(itm,start=i, end=i+1) for itm in mphon_lst]
        res.append("".join(syms))
    return res

def print_best(fst, num):
    bests_fst = fst.copy()
    bests_fst.n_best(num)
    bests_fst.minimize()
    raw_paths = bests_fst.extract_paths(output='raw')
    for raw_path in raw_paths:
        weight, pairs_tuple = raw_path
        alg_syms_lst = [alg_syms for alg_syms, out_sym in pairs_tuple]
        alg_syms_str = " ".join(alg_syms_lst)
        print(weight, alg_syms_lst)


def prefer_final_zeros(raw_paths):
    """Select the symbol pair sequence where the zeros are near the end

sym_lst_lst -- a list of results, each consisting of a list
    of symbols (already selected according to other criteria)

Returns a sequence of (single) symbols where the zeros occur near the
end.  This normalizes gemination and lengthening so that the latter
component is the one which alternates with a zero.
    """
    if cfg.verbosity >= 20:
        print("prefer_final_zeros(raw_paths):", raw_paths)
    max_path_len = max([len(p) for w, p in raw_paths])
    raw_path_lst = []
    for weight, pairs_tuple in raw_paths:
        if cfg.verbosity >= 20:
            print(weight, pairs_tuple)
        mphon_lst = [insym for insym, outsym in pairs_tuple]
        mphon_lst_len = len(mphon_lst)
        if cfg.verbosity >= 20:
            print(mphon_lst)
        i = 5 * mphon_lst_len
        weight += 20 * (mphon_lst_len - max_path_len)
        for mphon in mphon_lst:
            if "Ø" in mphon:
                weight = weight + i
            i -= 5
        raw_path_lst.append((weight, pairs_tuple))
    result_path_lst = sorted(raw_path_lst, key=lambda raw_path: raw_path[0])
    if cfg.verbosity == 20:
        print("result_path_lst:", result_path_lst)
    return result_path_lst


def word_to_fsa_with_zeros(word, target_len, zero="Ø"):
    """Insert zeros freeley to make the word into a fst of target_lenth words

word -- a string

target_len -- length of the strings accepted by the result fsa

Returns a fsa that accepts all strings of length target_length where some zeros have been added to make the word be target_len long.
"""
    word_len = grapheme.length(word)
    zeros_fst = fs.string_to_fsa(zero)
    zeros_fst.repeat_n(target_len - word_len)
    fst = fs.string_to_fsa(word)
    fst.shuffle(zeros_fst)
    fst.minimize()
    return fst


def align_words(word_lst, zero="Ø", extra_zeros=0, best_count=10):
    """Aligns a list of words

word_lst -- the list of words to be aligned

zero -- the symbol inserted as a mark for deletion or epenthesis

extra_zeros -- the maximun number of zeros to be added in the longest
    words (the shorter may have more)

best_count -- the maximum number of results to be returned (maybe less
    if no feasible results are found)

Returns a list of tuples where each tuple consists of a weight and a
list morphophonemes.
    """
    target_len = max([grapheme.length(x) for x in word_lst]) + extra_zeros
    fst = word_to_fsa_with_zeros(word_lst[0], target_len, zero)
    if cfg.verbosity >= 10:
        print("first fst with shuffled zeros:\n", fst)
    i = 1
    for word in word_lst[1:]:
        word_fsa = word_to_fsa_with_zeros(word, target_len, zero)
        fst.cross_product(word_fsa)
        fst = accum_input_labels(fst)
        if cfg.verbosity >= 10:
            print("fst accumulated with", word, ":\n", fst)

    fst.n_best(best_count)
    fst.minimize()
    if cfg.verbosity >= 10:
        print("aligned result:\n", fst)
    weight_path_lst = fst.extract_paths(output='raw')
    weight_mphon_lst_lst = []
    for weight, path_lst in weight_path_lst:
        mphon_lst = [insym for insym, outsym in path_lst]
        weight_mphon_lst_lst.append((weight, mphon_lst))
    if cfg.verbosity >= 10:
        print("weight_mphon_lst_lst:", weight_mphon_lst_lst)
    return weight_mphon_lst_lst


def adjustment(mphon_lst):
    """Computes an context based adjustment of a result

mphon_lst -- a list of morphophonemes from the result of align_words

Returns a number to be added to the weight.
    """
    adj = 10 * len(mphon_lst)
    phon_set_lst = [set(mphon)  for mphon in mphon_lst]
    if phon_set_lst[-1] <= alphabet.vowel_set | set("Ø"):
        adj -= 20
    for this_set, next_set in zip(phon_set_lst[:-1], phon_set_lst[1:]):
        if len(this_set) == 1 and next_set == this_set | set("Ø"):
            adj -= 20
    return adj

def multialign(word_lst,
               zero="Ø",
               max_zeros=1, best_count=1):
    """Aligns a list of words according to similarity of their phonemes

word_lst -- a list of words (or morphs) to be aligned

zero -- the symbol that will mark deletions or epentheses

max_zeros -- maximum number of zeros to be inserted into the longest
    word

best_count -- number of results to returned

Returns a list of results each of which is a tuple of a weight and a
list of words aligned by inserting zeros in an optimal way.

    """
    if len(word_lst) < 1:
        return []                 # nothing to align
    elif len(word_lst) == 1:
        return ((0,(word_lst),),) # only one word to align

    # process only the subset of unique words starting from the longest
    unique_words_lst = sorted(list(set(word_lst)),
                              key= lambda word: -len(word))
    aligned_results_lst = []
    for i in range(max_zeros + 1):
        result_lst = align_words(unique_words_lst,
                                 zero=zero,
                                 extra_zeros=i,
                                 best_count=best_count+9)
    
        for weight, alg_syms_lst in result_lst[:best_count]:
            if weight >= 1000:
                continue
            if cfg.verbosity >= 10:
                print("alg_syms_lst", alg_syms_lst)
            weight += adjustment(alg_syms_lst)
        
            aligned_words_lst = list_of_aligned_words(alg_syms_lst)

            # apply the alignment to the original word list with possible duplicates
            d = {}
            for uniqw, aligw in zip(unique_words_lst, aligned_words_lst):
                d[uniqw] = aligw
            aligned_orig_words_lst = [d[word] for word in word_lst]
            aligned_results_lst.append((weight, aligned_orig_words_lst))
    
    if cfg.verbosity >= 10:
        print("aligned_results_lst:", aligned_results_lst)
    return sorted(aligned_results_lst, key = lambda r: r[0])[:best_count]
        
def main():

    version = cfg.timestamp(__file__)
    
    import argparse
    arpar = argparse.ArgumentParser(
        "twol-multialign",
        description="""Version {}\nAligns lists of words separated by
        a DELIMITER.  See
        https://pytwolc.readthedocs.io/en/latest/alignment.html for
        detailed instructions. """.format(version))
    arpar.add_argument(
        "alphabet",
        help="An alphabet definition file with features and similarity sets.")
    arpar.add_argument(
        "-d", "--delimiter",
        help="Separates the two cognates, default is ' '",
        default=" ")
    arpar.add_argument(
        "-l", "--layout",
        choices=["vertical","list","horizontal"],
        help="output layout",
        default="vertical")
    arpar.add_argument(
        "-c", "--comment-separator",
        help="""Comment separator. Comments in input after this
        character are just copied to output. Input words are then
        also copied to the end of comments. Default separator is ''
        i.e. no comments.  Comments come to the output only in
        horizontal layout.""",
        default="")
    arpar.add_argument(
        "-w", "--weights",
        help="print also the weight of each alignment."
        " Default is not to print."
        " Works only if a comment separator is also set.",
        action="store_true")
    arpar.add_argument(
        "-x", "--extra-zeros", default=0, type=int,
        help="number of extra zeros to be tried in alignment")
    arpar.add_argument(
        "-n", "--number",
        help="number of best results to be printed. Default is 1",
        type=int, default=1)
    arpar.add_argument(
        "-v", "--verbosity",
        help="Level of diagnostic information to be printed. "
        "Default is 0",
        type=int, default=0)

    args = arpar.parse_args()

    if args.verbosity:
        cfg.verbosity = args.verbosity

    init(args.alphabet, all_zero_weight=1000)

    separator = args.delimiter
    import sys
    for line in sys.stdin:
        if args.comment_separator:
            word_str, comm, comments = \
                line.strip().partition(args.comment_separator)
        else:
            word_str, comm, comments = line.strip(), "", ""
        if args.verbosity > 0:
            print(word_str, args.comment_separator, comm)
        word_lst =  word_str.strip().split(args.delimiter)

        aligned_results_lst = multialign(word_lst,
                                         zero="Ø",
                                         max_zeros=args.extra_zeros,
                                         best_count=args.number)
        if cfg.verbosity >= 10:
            print("aligned_results_lst:", aligned_results_lst)
        for aligned_result in aligned_results_lst:
            print_result(aligned_result,
                         comments,
                         args.weights,
                         layout=args.layout)
    return

if __name__ == "__main__":
    main()



"""Aligns morphs that occur in the example words by inserting zero symbols

© Kimmo Koskenniemi, 2017-2018. This is free software under the GPL 3 license.
"""

def main():

    import twol.cfg as cfg
    version = cfg.timestamp(__file__)
    
    import argparse
    argparser = argparse.ArgumentParser(
        "python3 words2zerofilled.py",
        description="Aligns a set of word forms with morph boundaries"\
        " Version {} ".format(version))
    argparser.add_argument(
        "input",
        default="ksk-seg-examp.csv",
        help="moprheme names and segmented example words as a CSV file")
    argparser.add_argument(
        "output",
        default="ksk-alig-examp.csv",
        help="example words plus zero-filled aligned forms as a CSV file")
    argparser.add_argument(
        "alphabet",
        default="alphabet-test.text",
        help="An alphabet definition which determines"\
        " the weights for morphophonemes")
    argparser.add_argument(
        "-s", "--morph-separator",
        default=".",
        help="Separator between morphs in the word form, default is '.'")
    argparser.add_argument(
        "-d", "--csv-delimiter",
        default=",",
        help="Delimiter between the fields")
    argparser.add_argument(
        "-n", "--name-separator",
        default=".",
        help="separator between morpheme names"\
        " in the morpheme list,, default is '.'")
    argparser.add_argument(
        "-z", "--zero-symbol",
        default="Ø",
        help="symbol to be inserted in word forms to align them")
    argparser.add_argument(
        "-x", "--extra-zeros", default=0, type=int,
        help="number of extra zeros to be tried in alighnment")
    argparser.add_argument(
        "-v", "--verbosity", default=0, type=int,
        help="level of diagnostic and debugging output")
    args = argparser.parse_args()

    import re
    import csv
    import collections
    import grapheme

    cfg.verbosity = args.verbosity
    
    # STEP 1:
    # Read in the segmented words and collect the allomorphs of each morpheme

    morphs_of_morpheme = {} 
    """A dict to which allomorphs of each morpheme are collected:
    morphs_of_morpheme[morpheme_name] == ordered list its unique allomorphs.
    """
    seg_example_list = []
    """A list to which of all example words are collected. 
    Each word is represented as a list of (morpheme,morph) pairs.
    """
    stem_name_set = set()
    """Set of stem morphemes i.e. names of stem morphemes.
    """
    csvfile = open(args.input)

    reader = csv.DictReader(csvfile,
                            delimiter=args.csv_delimiter,
                            skipinitialspace=True)
    i = 0
    morphs_of_morpheme = {}
    for row in reader:
        morpheme_list = row["MORPHEMES"].strip().split(args.name_separator)
        morph_list = row["MORPHS"].strip().split(args.morph_separator)
        if args.verbosity >= 25:
            print(row["MORPHEMES"])
            print(morpheme_list)
            print(row["MORPHS"])
            print(morph_list)
        i = i + 1
        if len(morpheme_list) != len(morph_list):
            print("** line", i, ":", row["MORPHEMES"],
                    "is incompatible with", row["MORPHS"])
            continue
        if not morpheme_list:
            continue
        stem_name_set.add(morpheme_list[0])
        name_morph_pair_lst = list(zip(morpheme_list, morph_list))
        if args.verbosity >= 10:
            print("name_morph_pair_lst", name_morph_pair_lst)
        seg_example_list.append(name_morph_pair_lst)
        for morpheme, morph in name_morph_pair_lst:
            if args.verbosity >= 10:
                print("morpheme, morph:", morpheme, morph)
            morph = morph.strip()
            if morpheme not in morphs_of_morpheme:
                morphs_of_morpheme[morpheme] = [morph]
            else:
                if morph not in morphs_of_morpheme[morpheme]:
                    morphs = morphs_of_morpheme[morpheme]
                    morphs.append(morph)
                    morphs_of_morpheme[morpheme] = morphs
    if args.verbosity >= 5:
        print("morphs_of_morpheme", morphs_of_morpheme)

    csvfile.close()

    print("-- STEP 1 COMPLETED (seg_example_list, stem_name_set,"
          " morphs_of_morpheme done)--")

    # STEP 2:
    # align the allomorphs of each morpheme

    import twol.cfg as cfg
    #cfg.all_zero_weight = 1.0

    import twol.multialign as multialign
    
    multialign.init(args.alphabet, all_zero_weight=1)

    alignments = {}
    """All aligned morphs. index: morpheme name, value: sequence of
    aligned symbols.  Each aligned symbol has as many characters as
    there are items in the sequence.
    """

    for morpheme in sorted(morphs_of_morpheme.keys()):
        morphs = morphs_of_morpheme[morpheme]
        if len(morphs) == 1 and len(morphs[0]) == 0:
            aligned_morphs_lst = []
        else:
            if args.verbosity >= 5:
                print("morphs:", morphs)
            aligned_results_lst = \
                multialign.multialign(morphs,
                                      max_zeros=args.extra_zeros,
                                      best_count=1)
            if aligned_results_lst:
                weight, aligned_morphs_lst = aligned_results_lst[0]
            else:
                aligned_morphs_lst = []
        if args.verbosity >= 5:
            print("aligned_results_lst:", aligned_results_lst)
        alignments[morpheme] = aligned_morphs_lst

    print("-- STEP 2 COMPLETED (alignments done) --")

    # STEP 3:
    # Compute the zero filled morphs out of the sequences of aligned symbols

    aligned_morphs = {}
    """index: (morpheme, morph), value: zero-filled morph
    """

    for morpheme, aligned_morphs_lst in alignments.items():
        # e.g. "KOTA", ['kota', 'koda', 'kotØ', 'kodØ']
        if args.verbosity >= 5:
            print("aligned_morphs_lst:", aligned_morphs_lst)
        if morpheme not in aligned_morphs:
            aligned_morphs[morpheme] = collections.OrderedDict()
        if aligned_morphs_lst:
            original_morphs = [x.replace("Ø", "") for x in aligned_morphs_lst]
            for origm, zerofm in zip(original_morphs, aligned_morphs_lst):
                #if origm:
                #    aligned_morphs[morpheme][origm] = zerofm
                aligned_morphs[morpheme][origm] = zerofm
        else:
            aligned_morphs[morpheme] = {"": ""}
    if args.verbosity >= 5:
        print("aligned_morphs", aligned_morphs)

    print("-- STEP 3 COMPLETED (aligned_morphs done) --")

    # STEP 4:
    # Write the example word forms plus their a zero filled morphs

    out_file = open(args.output, "w", newline="")
    writer = csv.DictWriter(out_file,
                            ["MORPHEMES","MORPHS","ZEROFILLED"],
                            delimiter=args.csv_delimiter)
    forms_of_morphs = {}

    writer.writeheader()
    d = {}
    for seg_example in seg_example_list:
        if args.verbosity >= 20:
            print("seg_example:", seg_example)
        morpheme_lst = [morpheme for morpheme, morph in seg_example]
        morph_lst = [morph for morpheme, morph in seg_example]
        zero_filled_morph_lst = \
            [aligned_morphs[morpheme].get(morph.replace("Ø", ""), "")
             for (morpheme, morph) in seg_example]
        if args.verbosity >= 20:
            print("zero_filled_morph_lst:", zero_filled_morph_lst)
        d["MORPHEMES"] = args.name_separator.join(morpheme_lst)
        d["MORPHS"] = args.morph_separator.join(morph_lst)
        d["ZEROFILLED"] = args.morph_separator.join(zero_filled_morph_lst)
        writer.writerow(d)
        if morph_lst[0] not in forms_of_morphs:
            forms_of_morphs[morph_lst[0]] = set()
        forms_of_morphs[morph_lst[0]].add(" ".join(x for x in morpheme_lst[1:]))

    print("-- STEP 4 COMPLETED (zero-filled morphs and the CSV file done) --")
    return

if __name__ == "__main__":
    main()



__all__ = [
    "cfg",
    "fs",
    "twbt",
    "twexamp",
    "twolcomp",
    "twparser",
    "twrule",
    "discover",
    "aligner",
    "alphabet",
    "metric",
    "multialign",
    "raw2named",
    "table2words",
    "words2zerofilled",
    "zerofilled2raw"
]



""" aligner.py

Aligns two words (or morphs) by adding some zero symbols so that
phonemes in corresponding positions are optimally similar.

Copyright 2017-2019, Kimmo Koskenniemi

This is free software according to GNU GPL 3 license.
"""

import hfst
import twol.cfg as cfg
import twol.multialign as multialign

def align_two_words(in_word, out_word, aligner_fst, zero, number):
    w1 = hfst.fst(in_word)
    w1.insert_freely((zero,zero))
    w1.minimize()
    ###print(w1)

    w2 = hfst.fst(out_word)
    w2.insert_freely((zero,zero))
    w2.minimize()
    ###print(w2)

    w3 = hfst.HfstTransducer(w1)
    w3.compose(aligner_fst)
    w3.compose(w2)
    ###print(w1)

    w3.n_best(number)
    w3.minimize()
    ###print(w3)

    raw_paths = w3.extract_paths(output='raw')
    if cfg.verbosity >= 10:
        print("raw_paths:", raw_paths)
    return raw_paths

def print_result(raw_path, comments, weights, layout="horizontal"):

    weight, sym_pair_lst = raw_path
    
    if layout == "horizontal":
        mphon_lst = \
            [insym if insym == outsym else insym + ":" + outsym for insym, outsym in sym_pair_lst]
        mphonemic_str = " ".join(mphon_lst)
        if weights:
            print(mphonemic_str.ljust(40), weight)
        else:
            print(mphonemic_str)
        return
    if layout == "vertical" or layout == "list":
        in_word_lst = [insym for insym, outsym in sym_pair_lst]
        in_word_str = "".join(in_word_lst)
        out_word_lst = [outsym for insym, outsym in sym_pair_lst]
        out_word_str = "".join(out_word_lst)
    if layout == "vertical":
        print(in_word_str)
        print(out_word_str)
        if weights:
            print(29 * " ", weight)
        print()
    elif layout == "list":
        print(in_word_str, out_word_str)
        if weights:
            print(29 * " ", weight)
    return

def main():

    version = cfg.timestamp(__file__)
    
    import argparse
    arpar = argparse.ArgumentParser(
        "twol-aligner",
        description="""Aligns pairs of words separated by a
        colon. See https://pytwolc.readthedocs.io/en/latest/alignment.html
        for detailed instructions. Version {}""".format(version))
    arpar.add_argument(
        "metrics",
        help="FST computed with twol-metric from an alphabet file."
        " The FST contains weights for phoneme correspondences.")
    arpar.add_argument(
        "-d", "--delimiter",
        help="Separates the two cognates, default is ' '",
        default=" ")
    arpar.add_argument(
        "-l", "--layout",
        choices=["vertical","list","horizontal"],
        help="output layout",
        default="vertical")
    arpar.add_argument(
        "-c", "--comment-separator",
        help="""Comment separator. Comments in input after this
        character are just copied to output. Input words are then
        also copied to the end of comments. Default separator is ''
        i.e. no comments.  Comments come to the output only in
        horizontal layout.""",
        default="")
    arpar.add_argument(
        "-w", "--weights",
        help="print also the weight of each alignment."
        " Default is not to print."
        " Works only if a comment separator is also set.",
        action="store_true")
    arpar.add_argument(
        "-n", "--number",
        help="number of best results to be printed. Default is 1",
        type=int, default=1)
    arpar.add_argument(
        "-v", "--verbosity",
        help="Level of diagnostic information to be printed. "
        "Default is 0",
        type=int, default=0)

    args = arpar.parse_args()
    cfg.verbosity = args.verbosity

    algfile = hfst.HfstInputStream(args.metrics)
    aligner_fst = algfile.read()

    separator = args.delimiter
    import sys
    import twol.multialign as multialign
    
    for line in sys.stdin:
        if args.comment_separator:
            pair, comm, comments = \
                line.strip().partition(args.comment_separator)
        else:
            pair, comm, comments = line.strip(), "", ""
        if args.verbosity > 0:
            print(pair, args.comment_separator, comm)
        in_word, sep, out_word =  pair.strip().partition(args.delimiter)
        if not out_word:
            out_word = in_word

        raw_paths = align_two_words(in_word, out_word,
                                    aligner_fst,
                                    "Ø",
                                    args.number)
        for aligned_result in raw_paths:
            print_result(aligned_result,
                         comments,
                         args.weights,
                         layout=args.layout)


    return

if __name__ == "__main__":
    main()



"""Aligns multiple morphs according to phonological features
by adding zero symbols.

Copyright 2015-2020, Kimmo Koskenniemi

This is free software according to GNU GPL 3 License.
"""

import re, sys
import hfst
import grapheme
import twol.cfg as cfg
import twol.alphabet as alphabet

def remove_bad_transitions(fsa):
    """Copy the FSA excluding transitions with consonants and vowels"""
    old_bfsa = hfst.HfstBasicTransducer(fsa)
    new_bfsa = hfst.HfstBasicTransducer()
    for state in old_bfsa.states():
        new_bfsa.add_state(state)
        if old_bfsa.is_final_state(state):
            new_bfsa.set_final_weight(state, 0.0)
        for arc in old_bfsa.transitions(state):
            in_sym = arc.get_input_symbol()
            ##print(in_sym, "is")#####
            if alphabet.mphon_is_valid(in_sym):
                ##print("valid")#####
                target_st = arc.get_target_state()
                new_bfsa.add_transition(state, target_st, in_sym, in_sym, 0)
    result_fsa = hfst.HfstTransducer(new_bfsa)
    result_fsa.minimize()
    if cfg.verbosity >= 20:
        print("remove_bad_transitions:")
        print(result_fsa)
    return result_fsa

def shuffle_with_zeros(string, target_length):
    """Return a fsa where zeros are inserted in all possible ways
    
    string -- the string to which zero symbols are inserted

    target_length -- how long the strings after insertions must be

    Returns a fsa which accepts all the strings with the inserted zeros.
    All strings have exactly target_length symbols.
    """
    from twol.fs import string_to_fsa
    
    ### result_fsa = hfst.fst(string) # not correct for composed graphemes !!!
    result_fsa = string_to_fsa(string)
    l = grapheme.length(string)
    if l < target_length:
        n = target_length - l
        n_zeros_fsa = hfst.regex(" ".join(n * "Ø"))
        result_fsa.shuffle(n_zeros_fsa)
    result_fsa.minimize()
    result_fsa.set_name(string)
    if cfg.verbosity >= 30:
        print("shuffle_with_zeros:")
        print(result_fsa)
    return result_fsa

def set_weights(fsa):
    """Sets weights to transitions using mphon_weight()
    """
    bfsa = hfst.HfstBasicTransducer(fsa)
    for state in bfsa.states():
        for arc in bfsa.transitions(state):
            tostate = arc.get_target_state()
            insym = arc.get_input_symbol()
            outsym = arc.get_output_symbol()
            w = alphabet.mphon_weight(insym)
            arc.set_weight(w)
    weighted_fsa = hfst.HfstTransducer(bfsa)
    if cfg.verbosity >=20:
        print("set_weights:\n", weighted_fsa)
    return weighted_fsa

def multialign(strings, target_length):
    """Align a list of strings by making them target_lenght long
    
    Zero symbols are added optimally so that the sets of corresponding
    phonemes are similar.  Note that the alignment need not be feasible
    if the target lenght is too small and also that there may be
    all-zero correspondences if the target length is too long.
    """
    s1 = strings[0]
    fsa = shuffle_with_zeros(s1, target_length)
    for string in strings[1:]:
        suf_fsa = shuffle_with_zeros(string, target_length)
        fsa.cross_product(suf_fsa)      # results in a transducer
        if cfg.verbosity >=30:
            print("fsa\n", fsa)
        prod_fsa = hfst.fst_to_fsa(fsa)      # encodes the fst as a fsa
        fsa = remove_bad_transitions(prod_fsa)
        fsa.minimize()
    wfsa = set_weights(fsa)
    if cfg.verbosity >=20:
        print("multialign:\n", wfsa)
    return wfsa

def list_of_aligned_words(sym_lst):
    if not sym_lst:
        return []
    l = grapheme.length(sym_lst[0])
    res = []
    for i in range(l):
        syms = [grapheme.slice(itm,start=i, end=i+1) for itm in sym_lst]
        res.append("".join(syms))
    return res

def prefer_final_zeros(sym_lst_lst):
    """Select the symbol pair sequence where the zeros are near the end

    sym_lst_lst -- a list of results, each consisting of a list
    of symbols (already selected according to other criteria)

    Returns a sequence of (single) symbols where the zeros occur near
    the end.  This normalizes gemination and lengthening so that the
    latter component is the one which alternates with a zero.
    """
    best_bias = -1
    for sym_lst in sym_lst_lst:
        lst = [isym for isym in sym_lst]
        bias = 0
        i = 0
        for isym in lst:
            bias = bias + i * isym.count("Ø")
            i = i + 1
        #print("  ".join(lst), w, bias) ##
        if bias > best_bias:
            best_bias = bias
            best_sym_lst = lst
    return best_sym_lst

def classify_sym(sym):
    #from twolalign.alphabet import consonant_set
    char_set = set(sym)
    if char_set <= alphabet.consonant_set:
        if "Ø" in char_set:
            return "c"
        else: return "C"
    elif "Ø" in char_set:
        return "v"
    else: return "V"

def prefer_syl_struct(results):
    """Selects alignments according to syllable structure and zero count

    results -- list of tuples (weight, sym_pair_seq) (out of which the
    result list is chosen).  The sym_pair_seq is in the format that
    hfst.extract_paths() produces.

    Returns a list the best alternatives, i.e. those getting the lowest
    scores of the sum of syllable count and the number of instances
    where the former component of a CC or VV corresponds to zero.
    Elements in the returned list are sequences of symbols.
    """
    best_weight = results[0][0]
    best_bias = 99999
    best_lst = []
    for weight, sym_pair_seq in results:
        if weight > best_weight: break
        sym_lst = [isym for isym,outsym in sym_pair_seq]
        #print("sym_lst:", "  ".join(sym_lst)) ##
        csym_lst = [classify_sym(sym) for sym in sym_lst]
        csym_str = "".join(csym_lst)
        #print("csym_lst:", "  ".join(csym_lst)) ##
        syl_bias = len(re.findall(r"(C|c)+|(V|v)+", csym_str))
        #print("syl_bias:", syl_bias)###
        zero_bias = len(re.findall(r"(cC|vV)", csym_str))
        #print("zero_bias:", zero_bias)###
        bias = 3 * syl_bias + 2 * zero_bias
        if bias < best_bias:
            best_bias = bias
            best_lst = []
        if bias <= best_bias:
            best_lst.append(sym_lst)
            
    #print("best:", best_lst, "\n")####
    return best_lst

def aligner(words, max_zeros_in_longest, line):
    """Aligns a list of words according to similarity of their phonemes

    words -- a list of words (or morphs) to be aligned

    max_zeros_in_longest -- maximum number of zeros to be inserted into
    the longest word

    line -- the input line (used only in warning messages)

    cfg.all_zero_weight -- if phoneme set is {"Ø"} (default 100.0)

    Returns the best alignment as a list of raw morphophoneme.
    """

    if max([len(w) for w in words]) == 0:
        return []

    max_length = max([grapheme.length(x) for x in words])
    weighted_fsa = hfst.empty_fst()
    for m in range(max_length, max_length + max_zeros_in_longest + 1):
        R = multialign(words, m)
        if R.compare(hfst.empty_fst()):
            if cfg.verbosity > 1:
                print("target length", m, "failed")
            continue
        weighted_fsa.disjunct(R)
        weighted_fsa.minimize()
    weighted_fsa.n_best(10) 
    weighted_fsa.minimize() # accepts 10 best results
    results = weighted_fsa.extract_paths(output="raw")
    if cfg.verbosity >= 5:
        for w, sym_pair_seq in results:
            lst = [isym for isym, outsym in sym_pair_seq]
            mpw = ["{}::{:.2f}".format(x, alphabet.mphon_weight(x)) for x in lst]
            print(" ".join(mpw), "total weight = {:.3f}".format(w))
    if len(results) < 1:
        print("*** NO ALIGNMENTS FOR:", line, "***", results)
        return([])
    best_syl_struct = prefer_syl_struct(results)
    if cfg.final:
        best = prefer_final_zeros(best_syl_struct)
    else:
        best = best_syl_struct[0]
    return best

def main():
    import re, sys
    import hfst
    import grapheme

    version = "2020-02-01"
    import argparse
    arpar = argparse.ArgumentParser(
        "twol-multialign",
        description="Aligns sets of words separated by blanks"\
        " by inserting Ø symbols to make them equal length so that"\
        " similar phonemes correspond to each other. See"\
        " https://pytwolc.readthedocs.io/en/latest/alignment.html"\
        " for detailed instructions. Version {}".format(version))
    arpar.add_argument("alphabet",
        help="A file which defines the phonemes through their distinctive features",
        default="")
    arpar.add_argument("-l", "--layout",
        choices=["vertical","list","horizontal"],
        help="Output layout",
        default="vertical")
    arpar.add_argument("-f", "--final",
        help="Prefer deletion at the end",
        action="store_false")
    arpar.add_argument("-w", "--weights",
        help="Print the weight of each alignment",
        action="store_true")
    arpar.add_argument("-c", "--comments",
        help="Copy input words to the output lines as comments",
        default=False,
        action="store_true")
    arpar.add_argument("-v", "--verbosity",
        help="Level of diagnostic output",
        type=int, default=0)
    arpar.add_argument("-x", "--extra-zeros",
        help="Number of extra zeros allowed beyond the minimum",
        type=int, default=1)
    args = arpar.parse_args()

    cfg.verbosity = args.verbosity

    alphabet.read_alphabet(args.alphabet)

    if args.final:
        cfg.final = args.final
    valid_letters = alphabet.vowel_set | alphabet.consonant_set | set(" ")

    for line in sys.stdin:
        line = line.strip()
        lst = line.split("!", maxsplit=1)
        if len(lst) > 1:
            line = lst[0].strip()
            number = lst[1].strip() + " "
        else:
            number = ""
        if set(line) - valid_letters:
            print("** SOME LETTERS NOT VALID:",
                  set(line) - valid_letters,
                  "ON LINE: ", line)
            continue
        words = line.split()
        comment = number + " ".join(words)
            
        best = aligner(words, args.extra_zeros, line) # returns a list of morphoponemes
        if args.weights:
            weight = 0
            for mphon in best:
                weight += alphabet.mphon_weight(mphon)
            comment = "{:.2f} ".format(weight) + comment

        if args.layout == "horizontal":
            mphonemic_best = []
            for cc in best:
                grapheme_list = list(grapheme.graphemes(cc))
                lab = grapheme_list[0] if len(set(grapheme_list)) == 1 else cc
                mphonemic_best.append(lab)
            if args.comments:
                print(" ".join(mphonemic_best).ljust(40),"!", comment)
            else:
                print(" ".join(mphonemic_best))
        elif args.layout == "vertical":
            #print("best:", best) ###
            #print("len(best):", [len(x) for x in best]) ###
            print("\n".join(list_of_aligned_words(best)))
            print()
        elif args.layout == "list":
            print(" ".join(list_of_aligned_words(best)))
    return

if __name__ == "__main__":
    main()




"""twol-tester tests a set of two-level rules against examples

This program is useful for checking complete rule sets where all morphophonemes are constrained by some rules.  The twol-comp rule compiler checks indidual rules as it compiles them, so this program becomes useful when initial versions all rules have been written.  The formulae::

    lost = examples & rule[1] & rule[2] & ... & rule[n]

    neg = [pos.u .o. pairs*] - examples

    wrong = neg & rule[1] & rule[2] & ... & rule[n]

"""
import os
import re
import hfst
import twol.cfg as cfg
import twol.twexamp as twexamp
import twol.twbt as twbt

def paths(heading, fst):
    if cfg.verbosity >= 10:
        print(heading + "\n" + fst.extract_paths(max_cycles=1, output="text"))
    return

def main():
    import argparse
    arpar = argparse.ArgumentParser(
        "twol-tester",
        description = """A program for testing complete sets of twol rules""")
    arpar.add_argument(
        "-e", "--examples", action='store', nargs='+',
        help="""Either one name of a FST file that contains the examples or
            a list of names of files which contain the PSTR form examples
            used for compiling the rules.""",
        default=[None])
    arpar.add_argument(
        "-r", "--rules", action='store', nargs='+',
        help="""One or more files which contain the compiled rules as FSTs.
            The set  of rules ought to cover (almost) all morphophonemes.""",
        default=[None])
    arpar.add_argument(
        "-l", "--lost",
        help="""an FST that lists positive examples that are not
            accepted by all rules""",
        default="")
    arpar.add_argument(
        "-w", "--wrong",
        help="""an FST that lists the negative examples that are
            accepted gy by all rules""",
        default="")
    arpar.add_argument(
        "-v", "--verbosity",
        help="level of  diagnostic output",
        type=int, default=0)

    args = arpar.parse_args()
    cfg.verbosity = args.verbosity
    #
    # Build the FST of the example pair strings and
    # store in cfg.examples_fst and the cfg.input_symbol_set,
    # cfg.output_symbol_set, cfg.symbol_pair_set and
    # cfg.all_pairs_set
    #
    if len(args.examples) and args.examples[0].endswith(".fst"):
        twexamp.read_fst(args.examples)
    elif len(args.examples) > 0:
        twexamp.read_examples(args.examples, build_fsts=True)
    else:
        error("ERROR IN EXAMPLE FILE NAMES: {}".format(args.examples))
        #
    # Read in the compiled twol rule FST or FSTs
    #
    rule_fst_lst = []
    for rule_file_name in args.rules:
        if (not os.path.isfile(rule_file_name) and
            rule_file_name.endswith(".fst")):
            exit("RULE FST FILE {} DOES NOT EXIST",format(rule_file))
        fst_stream = hfst.HfstInputStream(rule_file_name)
        while not fst_stream.is_eof():
            fst = fst_stream.read()
            rule_fst_lst.append(fst)
        fst_stream.close()
    #
    # Build positive and negative examples 
    #
    pos_fst = cfg.examples_fst.copy()
    neg_fst = pos_fst.copy()
    paths("positive examples", neg_fst)
    neg_fst.input_project()
    pistar_fst = cfg.all_pairs_fst.copy()
    pistar_fst.repeat_star()
    neg_fst.compose(pistar_fst)
    neg_fst.minimize()
    neg_fst.subtract(pos_fst)
    paths("negative examples", neg_fst)
    #
    # Lost and wrong examples
    #
    if args.lost:
        remain_fst = pos_fst.copy()
        for fst in rule_fst_lst:
            remain_fst.intersect(fst)
        lost_fst = pos_fst.copy()
        lost_fst.subtract(remain_fst)
        lost_stream = hfst.HfstOutputStream(filename=args.lost)
        lost_stream.write(lost_fst)
        lost_stream.close()
    if args.wrong:
        wrong_fst = neg_fst.copy()
        for fst in rule_fst_lst:
            wrong_fst.intersect(fst)
        wrong_stream = hfst.HfstOutputStream(filename=args.wrong)
        wrong_stream.write(wrong_fst)
        wrong_stream.close()

if __name__ == "__main__":

    main()



"""A script for converting zero-filled examples into examples with raw morphophonemes

Raw morphophonemes are named according to a set of principal forms
(principal parts) which are assumed to reflect all morphophonemic
alternations in a lexeme.  The script assumes that the examples within
a lexeme occur all in the same order as for their form.

The code still relies on the assumption that the stem comes as the
first morph (and thus some minor modifications would be required in
order to handle prefixing languages).

© Kimmo Koskenniemi, 2018. This is free software under GPL 3 license.

"""

def main():

    import twol.cfg as cfg
    version = cfg.timestamp(__file__)
    
    import argparse
    argparser = argparse.ArgumentParser(
        "python3 zerofilled2raw.py",
        description="Forms raw morphophonemes out of zero-filled"\
        " morphs and produces a space-separated pair string"\
        " representation for the word suitable for"\
        " twol-comp or twol-discov. Version {}".format(version))
    argparser.add_argument(
        "input",
        help="zero-filled example words as a CSV file")
    argparser.add_argument(
        "output",
        help="The output file in CSV format with a new column"\
        " where the words are represented with raw"\
        " morhpophonemes from zero-filling.")
    argparser.add_argument(
        "affix_info",
        help="Principal forms and morphophonemic affixes as a CSV file")
    argparser.add_argument(
        "-d", "--csv-delimiter",
        default=",",
        help="Delimiter between the fields, default=','")
    argparser.add_argument(
        "-s", "--morph-separator",
        default=".",
        help="Separator between morphs in the word form, default='.'")
    argparser.add_argument(
        "-n", "--name-separator",
        default=".",
        help="Separator between morpheme names in the morpheme list")
    argparser.add_argument(
        "-z", "--zero-symbol",
        default="Ø",
        help="Symbol inserted in word forms to align them")
    argparser.add_argument(
        "-v", "--verbosity",
        default=0,
        type=int,
        help="level of diagnostic and debugging output")
    args = argparser.parse_args()

    import re
    import csv
    import collections

    principal_lst = []
    """"List of principal forms or principal parts, i.e. the forms which
    uniquely determine the morphophonemic variations that may occur
    within the stem.

    """
    feat2mphons = {}

    # Read in the feature combinations of principal forms and
    # the morphophonemic representations of affix features
    with open(args.affix_info, "r") as afffil:
        affrdr = csv.reader(afffil,
                            delimiter=args.csv_delimiter,
                            skipinitialspace=True)
        for row in affrdr:
            if row[1] == '+':
                feat = row[0]
                if feat not in principal_lst:
                    principal_lst.append(feat)
            else:
                feat2mphons[row[0]] = row[1]
    if args.verbosity >= 10:
        print("principal_lst =", principal_lst)####
        print("feat2mphons =", feat2mphons)####

    # Read in the morpheme names and the zero-filled morphs

    stem_morpheme_data = collections.OrderedDict()
    """Indexed by stem morpheme name, value is a list of the original data
    for that stem morpheme.  Each value consists of a tuple of fields
    (MORPHEMES, MORPHS, ALIGNED) in the original data.
    """
    with open(args.input, "r") as infil:
        rdr = csv.DictReader(infil,
                             delimiter=args.csv_delimiter,
                             skipinitialspace=True)
        for row in rdr:
            names = row["MORPHEMES"].strip()
            orig_morphs = row["MORPHS"].strip()
            zerof_morphs = row["ZEROFILLED"].strip()
            if (not names) or (not zerof_morphs):
                continue
            name_lst = names.split(args.name_separator, maxsplit=1)
            stem_name = name_lst[0]
            form_name = ".".join(name_lst[1:]) if len(name_lst) > 1 else ""
            zerof_morph_lst = zerof_morphs.split(args.morph_separator,
                                                 maxsplit=1)
            if stem_name not in stem_morpheme_data:
                stem_morpheme_data[stem_name] = []
            stem_morpheme_data[stem_name].append((form_name,
                                                  orig_morphs,
                                                  zerof_morph_lst))

    ofil = open(args.output, "w")
    writer = csv.DictWriter(ofil, fieldnames=["MORPHEMES", "MORPHS",
                                              "ZEROFILLED", "RAW"])
    writer.writeheader()

    for stem_morpheme, data_lst in stem_morpheme_data.items():
        princ_zstem_lst =[]
        if args.verbosity >= 10:
            print("*** stem_morpheme, data_lst:", stem_morpheme, data_lst)
        # select the principal forms of this stem morpheme
        for data in data_lst:
            form_name, orig_morphs, zerof_morph_lst = data
            if form_name in principal_lst:
                princ_zstem_lst.append(zerof_morph_lst[0])
        # form the raw morphophonemes by combining corresponding
        # symbols
        if args.verbosity >= 10:
            print("*** princ_zstem_lst:", princ_zstem_lst) ###
        lgth = len(princ_zstem_lst[0])
        zstem_rawsym_lst = []
        for i in range(lgth):
            lst = []
            for princ_zstem in princ_zstem_lst:
                lst.append(princ_zstem[i])
                # print(stem_morpheme, i, lst)###
            raw_seq = "".join(lst)
            if re.match(r"^(.)(\1)*$", raw_seq):
                raw_sym = raw_seq[0]  # abbreviate if all identical
            else:
                raw_sym = "{" + raw_seq + "}"
            zstem_rawsym_lst.append(raw_sym)
        zstem_pairsym_str = " ".join(zstem_rawsym_lst)
        # Output the data augmented with the representation with raw
        # morphophonemes
        for data in data_lst:
            form_name, orig_morphs, zerof_morph_lst = data
            form_part = args.name_separator + form_name if form_name else ""
            row["MORPHEMES"] = (stem_morpheme + form_part).strip() 
            row["MORPHS"] = orig_morphs
            orig_zerof_morphs = args.morph_separator.join(zerof_morph_lst)
            row["ZEROFILLED"] = orig_zerof_morphs
            raw_lst = [zstem_pairsym_str]
            feat_lst = form_name.split(args.name_separator)
            for feat in feat_lst:
                raw_lst.append(feat2mphons[feat])
            row["RAW"] = " ".join(raw_lst)
            writer.writerow(row)
    return

if __name__ == "__main__":
    main()



"""metric.py

Produces a weighted finite-state transducer (WFST) out of the alphabet
definition as interpreted by alphabet.py.  The WFST is the used by the
aligner.py program.

Copyright 2020, Kimmo Koskenniemi

This is free software according to GNU GPL 3 license.

"""

import sys
import re
import twol.alphabet as alphabet
import twol.cfg as cfg
import hfst

within_set_lst = []
forall_lst = []

pair_weight_dict = {}

def alignment_fst():
    pair_weight_lst = []
    for insym in alphabet.consonant_set:
        mphon = "Ø" + insym
        if alphabet.mphon_is_valid(mphon):
            xyw = "{}:{}::{}".format(insym, "Ø", alphabet.mphon_weight(mphon))
        pair_weight_lst.append(xyw)
        mphon = insym + "Ø"
        if alphabet.mphon_is_valid(mphon):
            xyw = "{}:{}::{}".format("Ø", insym, alphabet.mphon_weight(mphon))
        pair_weight_lst.append(xyw)
        for outsym in alphabet.consonant_set:
            mphon = insym + outsym
            if alphabet.mphon_is_valid(mphon):
                xyw = "{}:{}::{}".format(insym, outsym,
                                         alphabet.mphon_weight(mphon))
                pair_weight_lst.append(xyw)
    for insym in alphabet.vowel_set:
        mphon = "Ø" + insym
        if alphabet.mphon_is_valid(mphon):
            xyw = "{}:{}::{}".format(insym, "Ø", alphabet.mphon_weight(mphon))
        pair_weight_lst.append(xyw)
        mphon = insym + "Ø"
        if alphabet.mphon_is_valid(mphon):
            xyw = "{}:{}::{}".format("Ø", insym, alphabet.mphon_weight(mphon))
        pair_weight_lst.append(xyw)
        for outsym in alphabet.vowel_set:
            mphon = insym + outsym
            if alphabet.mphon_is_valid(mphon):
                xyw = "{}:{}::{}".format(insym, outsym,
                                         alphabet.mphon_weight(mphon))
                pair_weight_lst.append(xyw)
    pair_weight_str = "|".join(pair_weight_lst)
    if cfg.verbosity >= 20:
        print("\npair_weight_str:", pair_weight_str)

    loop_sets = {}
    loop_sets["Consonants"] = alphabet.consonant_set - {"Ø"}
    loop_sets["Vowels"] = alphabet.vowel_set- {"Ø"}
    if cfg.verbosity >= 20:
        print("\nloop_sets:", loop_sets)
    for_all_lst = []
    for expr, var, loop_set in alphabet.for_definitions_lst:
        if cfg.verbosity >= 25:
            print(expr, var, loop_set)
        for x in loop_sets[loop_set]:
            if cfg.verbosity >= 25:
                print(x, var, loop_set)
            item_expr = re.sub(var, x, expr)
            if cfg.verbosity >= 25:
                print("item_expr:", item_expr)
            for_all_lst.append(item_expr)
    for_all_str = "|".join(for_all_lst)
    if cfg.verbosity >= 20:
        print("\nfor_all_str:", for_all_str)

    exceptions_str = " | ".join(alphabet.exception_lst)
    if cfg.verbosity >= 20:
        print("\nexceptions_str:", exceptions_str)

    fst = hfst.regex(pair_weight_str + "|" + for_all_str + "|" + exceptions_str)
    fst.repeat_star()
    fst.minimize()
    return fst

def main():
    #last_modified_date = datetime.fromtimestamp(mtime)
    
    version = cfg.timestamp(__file__)
    
    import argparse
    arpar = argparse.ArgumentParser(
        "twol-metric",
        description="""Builds a distance metric FST out of an alphabet
        description. See
        https://pytwolc.readthedocs.io/en/latest/alignment.html
        for detailed instructions. Version {}""".format(version)
    )
    arpar.add_argument(
        "alphabet",
        help="An alphabet definition with features and similarity sets")
    arpar.add_argument(
        "metrics",
        help="FST which contains weights for preferring alternative alignments")
    arpar.add_argument(
        "-v", "--verbosity",
        help="Level of diagnostic output printed, default=0",
        type=int, default=0)
    args = arpar.parse_args()

    cfg.verbosity = args.verbosity
    cfg.all_zero_weight = 1000

    alphabet.read_alphabet(args.alphabet)

    fst = alignment_fst()

    fstfile = hfst.HfstOutputStream(filename=args.metrics)
    fstfile.write(fst)
    fstfile.flush()
    fstfile.close()
    return

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import sys, re, json
from codecs import open
from pprint import pprint

import tatsu
from tatsu import compile
from tatsu.ast import AST
#from tatsu import ast
#from ast import AST
from tatsu.walkers import NodeWalker
#from tatsu import walkers
#from walkers import NodeWalker

from tatsu.exceptions import ParseException, FailedParse, ParseError, FailedSemantics

import hfst
import twol.cfg as cfg
import twol.twexamp as twexamp

class TwolRegexSemantics(object):

    def define(self, ast):
        cfg.definitions[ast.left] = ast.right
        return ("=", ast.left, ast.right)
    
    def identifier(self, ast):
        string = ast.token.strip()
        return string

    def right_arrow_rule(self, ast):
        result = ("=>", ast.left, ast.right)
        return result

    def left_arrow_rule(self, ast):
        result = ("<=", ast.left, ast.right)
        return result

    def double_arrow_rule(self, ast):
        result = ("<=>", ast.left, ast.right)
        return result

    def exclusion_rule(self, ast):
        result = ("/<=", ast.left, ast.right)
        return result

    def context_lst(self, ast):
        result = ast.left.extend(ast.right)
        return result

    def context(self, ast):
        lc = ast.left if ast.left else ""
        rc = ast.right if ast.right else ""
        result = [(lc, rc)]
        return result

    def union(self, ast):
        return "[{} | {}]".format(ast.left, ast.right)

    def intersection(self, ast):
        return "[{} & {}]".format(ast.left, ast.right)

    def difference(self, ast):
        return "[{} - {}]".format(ast.left, ast.right)

    def concatenation(self, ast):
        return "[{} {}]".format(ast.left, ast.right)

    def composition(self, ast):
        return "[{} .o. {}]".format(ast.left, ast.right)

    def crossproduct(self, ast):
        return "[{} .x. {}]".format(ast.left, ast.right)

    def Kleene_star(self, ast):
        return "[{}]*".format(ast.expr)

    def Kleene_plus(self, ast):
        return "[{}]+".format(ast.expr)

    def Upper(self, ast):
        return "[{}].u".format(ast.expr)

    def Lower(self, ast):
        return "[{}].l".format(ast.expr)

    def One_but_not(self, ast):
        return r"[PI - [{}]]".format(ast.expr)

    def optexpression(self, ast):
        return "({})".format(ast.expr)

    def subexpression(self, ast):
        return "[{}]".format(ast.expr)

    def symbol_or_pair(self, ast):
        string = ast.token.strip()
        failmsg = []
        pat = re.compile(r"""^
        (?P<up>[^{}:\s]*
         |
         \{[^{}:\s]+\})
        :
        (?P<lo>[^{}:\s]*)
        $""", re.X)
        m = re.match(pat, string)
        if m:                       # it is a pair with a colon
            up = m.group("up")
            up_quoted = re.sub(r"([{}])", r"%\1", up)
            lo = m.group("lo")
            if up and (up not in cfg.input_symbol_set):
                failmsg.append("input symbol '{}'".format(up))
            if lo and (lo not in cfg.output_symbol_set):
                failmsg.append("output symbol '{}'".format(lo))
            if up and lo and ((up, lo) not in cfg.symbol_pair_set):
                failmsg.append("symbol pair '{}'".format(string))
            if failmsg:
                cfg.error_message = " and ".join(failmsg) + " not in alphabet"
                raise FailedSemantics(cfg.error_message)
            elif up and lo:         # it is a valid pair with a colon
                return "{}:{}".format(up_quoted, lo)
            elif up and (not lo):
                return "[{} .o. PI]".format(up_quoted)
            elif (not up) and lo:
                return "[PI .o. {}]".format(lo)
            else:
                return "PI"
        m = re.fullmatch(r"[^{}\s:]+", string)
        if m:                       # its either a defined sym or a surf ch
            if string in cfg.definitions:
                return "{}".format(string)
            elif (string in cfg.output_symbol_set) and (string in cfg.input_symbol_set):
                return "{}:{}".format(string, string)
            elif string in {'BEGIN', 'END'}:
                return string
        cfg.error_message = "'" + string + "' is an invalid pair/definend symbol"
        raise FailedSemantics(cfg.error_message)

class TwolFstSemantics(object):

    def define(self, ast):
        expr_fst = ast.right.copy()
        def_name = ast.left
        cfg.definitions[def_name] = expr_fst
        return ("=", ast.left, ast.right)
    
    def identifier(self, ast):
        string = ast.token.strip()
        return string

    def right_arrow_rule(self, ast):
        result = ("=>", ast.left, ast.right)
        return result

    def output_coercion_rule(self, ast):
        result = ("<=", ast.left, ast.right)
        return result

    def input_coercion_rule(self, ast):
        result = ("<--", ast.left, ast.right)
        return result

    def double_arrow_rule(self, ast):
        result = ("<=>", ast.left, ast.right)
        return result

    def exclusion_rule(self, ast):
        result = ("/<=", ast.left, ast.right)
        return result

    def context_lst(self, ast):
        left_lst = ast.left.copy()
        right_lst = ast.right.copy()
        result = left_lst.copy()
        result.extend(right_lst)
        return result

    def context(self, ast):
        lc = ast.left.copy() if ast.left else hfst.epsilon_fst()
        rc = ast.right.copy() if ast.right else hfst.epsilon_fst()
        lc.substitute("END", "BEGIN")
        #print(lc)###
        #print(rc)###
        result = [(lc, rc)]
        return result

    def union(self, ast):
        name = "[{} | {}]".format(ast.left.get_name(), ast.right.get_name())
        result_fst = ast.left.copy()
        result_fst.disjunct(ast.right)
        result_fst.minimize()
        result_fst.set_name(name)
        return result_fst

    def intersection(self, ast):
        name = "[{} & {}]".format(ast.left.get_name(), ast.right.get_name())
        result_fst = ast.left.copy()
        result_fst.conjunct(ast.right)
        result_fst.minimize()
        result_fst.set_name(name)
        return result_fst

    def difference(self, ast):
        name = "[{} - {}]".format(ast.left.get_name(), ast.right.get_name())
        result_fst = ast.left.copy()
        result_fst.minus(ast.right)
        result_fst.minimize()
        result_fst.set_name(name)
        return result_fst

    def concatenation(self, ast):
        name = "[{} {}]".format(ast.left.get_name(), ast.right.get_name())
        result_fst = ast.left.copy()
        result_fst.concatenate(ast.right)
        result_fst.minimize()
        result_fst.set_name(name)
        return result_fst

    def composition(self, ast):
        name = "[{} .o. {}]".format(ast.left.get_name(), ast.right.get_name())
        result_fst = ast.left.copy()
        result_fst.compose(ast.right)
        result_fst.minimize()
        result_fst.set_name(name)
        return result_fst

    def crossproduct(self, ast):
        name = "[{} .x. {}]".format(ast.left.get_name(), ast.right.get_name())
        result_fst = ast.left.copy()
        result_fst.cross_product(ast.right)
        result_fst.set_name(name)
        return result_fst

    def Kleene_star(self, ast):
        name = "[{}]*".format(ast.expr.get_name())
        result_fst = ast.expr.copy()
        result_fst.repeat_star()
        result_fst.minimize()
        result_fst.set_name(name)
        return result_fst

    def Kleene_plus(self, ast):
        name = "[{}]+".format(ast.expr.get_name())
        result_fst = ast.expr.copy()
        result_fst.repeat_plus()
        result_fst.minimize()
        result_fst.set_name(name)
        return result_fst

    def Upper(self, ast):
        """Input projection

        Note that the result is outside the Pi*, i.e. not a valid
        two-level expression.  This operator should be avoided at all
        costs.
        """
        name = "[{}].u".format(ast.expr.get_name())
        result_fst = ast.expr.copy()
        result_fst.input_project()
        result_fst.set_name(name)
        return result_fst

    def Lower(self, ast):
        """Output projection

        Note that the result is outside the Pi*, i.e. not a valid
        two-level expression.  This operator should be avoided at all
        costs.
        """
        name = "[{}].l".format(ast.expr.get_name())
        result_fst = ast.expr.copy()
        result_fst.output_project()
        result_fst.minimize()
        result_fst.set_name(name)
        return result_fst

    def Morphophonemic(self, ast):
        """Surface completion

        Returns a FST which accepts sequences of valid pairs whose
        input side is accepted by the input side of the argument.
        For a single symbol pair k:g it is equivalent to k:
        """
        name = "[{}].m".format(ast.expr.get_name())
        result_fst = ast.expr.copy()
        result_fst.input_project()
        all_pairs_fst = cfg.all_pairs_fst.copy()
        result_fst.compose(all_pairs_fst)
        result_fst.minimize()
        result_fst.set_name(name)
        return result_fst

    def Surface(self, ast):
        """Morphophonemic completion

        Returns a FST which accepts sequences of valid pairs whose
        ouput side is accepted by the output side of the argument.
        For a single symbol pair k:g it is equivalent to :g
        """
        name = "[{}].s".format(ast.expr.get_name())
        temp_fst = ast.expr.copy()
        temp_fst.output_project()
        result_fst = cfg.all_pairs_fst.copy()
        result_fst.compose(temp_fst)
        result_fst.minimize()
        result_fst.set_name(name)
        return result_fst

    def One_but_not(self, ast):
        name = r"\[{}]".format(ast.expr.get_name())
        result_fst = cfg.all_pairs_fst.copy()
        result_fst.minus(ast.expr)
        result_fst.minimize()
        result_fst.set_name(name)
        return result_fst

    def optexpression(self, ast):
        result_fst = ast.expr.copy()
        name = result_fst.get_name()
        result_fst.optionalize()
        result_fst.minimize()
        result_fst.set_name("({})".format(name))
        return result_fst

    def subexpression(self, ast):
        name = "[{}]".format(ast.expr.get_name())
        result_fst = ast.expr.copy()
        result_fst.set_name(name)
        return result_fst

    def symbol_or_pair(self, ast):
        string = ast.token.strip()
        failmsg = []
        #pat = re.compile(r"""^
        #(?P<up>[a-zšžåäöüõA-ZÅÄÖ0-9'´`]*
        # |
        # \{[a-zåäöüõA-ZÅÄÖØ'´`]+\})
        #:
        #(?P<lo>[a-zšžåäöüõA-ZÅÄÖØ'´`]*)
        #$""", re.X)
        pat = re.compile(r"""^
        (?P<up>[^- \[\]\/<>=_;,.|&*+\\()\{\}]*
         |
         \{[^{}:\s]+\})
        :
        (?P<lo>[^- \[\]\/<>=_;,.|&*+\\()\{\}]*)
        $""", re.X)
        m = re.match(pat, string)
        if m:                       # it is a pair with a colon
            up = m.group("up")
            up_quoted = re.sub(r"([{}])", r"%\1", up)
            lo = m.group("lo")
            if up and (up not in cfg.input_symbol_set):
                failmsg.append("input symbol '{}'".format(up))
            if lo and (lo not in cfg.output_symbol_set):
                failmsg.append("output symbol '{}'".format(lo))
            if up and lo and ((up, lo) not in cfg.symbol_pair_set):
                failmsg.append("symbol pair '{}'".format(string))
            if failmsg:
                cfg.error_message = " and ".join(failmsg) + " not in alphabet"
                raise FailedSemantics(cfg.error_message)
            elif up and lo:         # it is a valid pair with a colon
                result_fst = hfst.regex(up_quoted + ':' + lo)
                result_fst.set_name(string)
                return result_fst
            elif up and (not lo):
                result_fst = hfst.regex(up_quoted)
                result_fst.compose(cfg.all_pairs_fst)
                result_fst.set_name(string)
                return result_fst
            elif (not up) and lo:
                result_fst = cfg.all_pairs_fst.copy()
                lo_fst = hfst.regex(lo)
                result_fst.compose(lo_fst)
                result_fst.set_name(string)
                return result_fst
            else:
                result_fst = cfg.all_pairs_fst.copy()
                result_fst.set_name("PI")
                return result_fst
        m = re.fullmatch(r"[^][|+*{}:\s]+", string)
        if m:                       # its either a defined sym or a surf ch
            if string in cfg.definitions:
                result_fst = cfg.definitions[string].copy()
                result_fst.set_name(string)
                return result_fst
            elif (string in cfg.output_symbol_set) and (string in cfg.input_symbol_set):
                result_fst =  hfst.regex(string)
                result_fst.set_name(string)
                return result_fst
            elif string in {'BEGIN', 'END'}:
                result_fst = hfst.regex(string)
                result_fst.set_name(string)
                return result_fst
        cfg.error_message = "'" + string + "' is an invalid pair/definend symbol"
        raise FailedSemantics(cfg.error_message)

    def boundary(self, ast):
        result_fst = hfst.regex("END")
        #print(result_fst)####
        result_fst.set_name(".#.")
        return result_fst

def init():
    """Initializes the module and compiles and returns a tatsu parser

    grammar_file -- the name of the file containing the EBNF grammar
    for rules

    """
    import os
    dir = os.path.dirname(os.path.abspath(__file__))
    grammar_file = dir + "/twolcsyntax.ebnf"
    grammar = open(grammar_file).read()
    parser = compile(grammar)
    return parser

def parse_rule(parser, line_nl, line_no, line_lst, start="expr_start"):
    """Parse one rule or definiton or any constituent given as start

    parser -- a tatsu parser which parses the EBNF grammar for two-level rules
    line_nl -- the string that contains the rule or definition to be parsed

    keyword arguments:
    start -- the element in the EBNF grammar where to start the parsing
"""
    line = line_nl.strip()
    if (not line) or line[0] == '!':
        return "!", None, None  # it was a comment or an empty line
    rulepat = r"^.* +(=|<=|=>|<=>|/<=|<--) +.*$"
    try:
        m = re.match(rulepat, line)
        if m:
            #print("groups:", m.groups())###
            if m.group(1) == '=':
                op, name, expr_fst = parser.parse(line, start='def_start',
                                                  semantics=TwolFstSemantics())
                return op, name, expr_fst
            elif m.group(1) in {'=>', '<=', '<=>', '/<=', '<--'}:
                op, x_fst, contexts = parser.parse(line, start='rul_start',
                                                   semantics=TwolFstSemantics())
                return op, x_fst, contexts
        else:
             return "?", None, None
    except ParseException as e:
        print("\n" + 40 * "*")
        print("ERROR WAS IN INPUT LINES:",
              line_no, "-", line_no + len(line_lst) - 1)
        print("".join(line_lst))
        print("THE ERROR IS LIKELY ABOVE THE '^' OR BEFORE IT")
        msg = str(e)
        lst = msg.split("\n")
        if len(lst) >= 3:
            print(lst[1])
            print(lst[2], "<---", e.__class__.__name__, "HERE")
            if cfg.error_message:
                print("EXPLANATION:")
                print("    ", cfg.error_message)
                cfg.error_message = ""
        else:
            print(e) ###
            print(str(e))
        print(40 * "*" + "\n")
        return "?", None, None

def main():
    import hfst
    import argparse
    import twol.twbt as twbt
    import twol.cfg as cfg
    import twol.twexamp as twexamp
    arpar = argparse.ArgumentParser(
        description="A compiler and tester for two-level rules")
    arpar.add_argument("start",
                        help="start parsing from",
                        default="expr_start")
    args = arpar.parse_args()
    twexamp.read_fst(filename="nounex.fst")
    parser = init()
    for line_nl in sys.stdin:
        line = line_nl.strip()
        #print(line)
        result = parser.parse(line, start=args.start,
                              semantics=TwolFstSemantics())
        if args.start == "def_start":
            op, left, right, source = result
            print(left, "=")
            twbt.ppfst(right)
        elif args.start == "rul_start":
            op, left, right, source = result
            twbt.ppfst(left)
            print(op)
            for lc, rc in right:
                twbt.ppfst(lc, title="left context")
                twbt.ppfst(rc, title="right context")
        elif args.start == "expr_start":
            fst = result
            #print(fst)
            twbt.ppfst(fst, True)
        elif op == "?":
            print("Incorrect: " + line)
        return

if __name__ == '__main__':
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="twol", # of the PyPI project and package
    version="0.6.5",
    author="Kimmo Koskenniemi",
    author_email="koskenni@gmail.com",
    description="Tools for simplified two-level morphology",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/koskenni/twol",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "twol-aligner = twol.aligner:main",
            "twol-metric = twol.metric:main",
            "twol-multialign = twol.multialign:main",
            "twol-table2words = twol.table2words:step1",
            "twol-words2zerofilled = twol.words2zerofilled:main",
            "twol-zerofilled2raw = twol.zerofilled2raw:main",
            "twol-raw2named = twol.raw2named:main",
            "twol-comp = twol.twolcomp:main",
            "twol-examples2fst = twol.twexamp:main",
            "twol-discov =  twol.discover:main",
            "twol-tester = twol.tester:main"
        ]
    },
    include_package_data=True,
    python_requires='>=3.6,<=3.7',
        install_requires=[
            'grapheme',
            'hfst',
            ##'hfst==3.15.0.0b0',
            'TatSu==4.4.0',
    ]

)



"""Proposes a rule for the distribution of one phoneme correspondence.

© Kimmo Koskenniemi, 2019. Free software under the GPL 3 or later.

A correspondence is a sequence of phonemes, e.g. "aaä" would be the
correspondence of phonemes "a", "a" and "ä".  The standard input
consists of cognate sets which are represented as a sequence of
correspondences. Correspondences where all phonemes are equal are
abbreviated as a single phoneme.

The command

 $ python3 multialign.py -l horizontal

can produce lines of such aligned cognate word forms, e.g.

  o r ji aØ

out of lines of space-separated cognate words, e.g.

  orja ori

"""

import twol.cfg as cfg

corr_width = 0                  # number of languages
corr_set = set()                # set of all phoneme correspondences
context_dict = {}               # key: corr, value: list of contexts
phoneme_set = set()             # set of all phonemes in the correspondences

from collections import deque

def read_aligned_cognates(file_name):
    global corr_width
    alcog_file = open(file_name, "r")
    left_queue = deque()
    right_queue = deque()
    for line_nl in alcog_file:
        line = line_nl.strip()
        if line.startswith("!"):
            continue
        lst = line.split("!", maxsplit=1)
        alcog_str = lst[0]
        corr_lst = alcog_str.strip().split()
        mx = max([len(c) for c in corr_lst])
        if mx > corr_width:
            corr_width = mx
        left_queue.clear()
        left_queue.append("#")
        right_queue.clear()
        right_queue.extend(corr_lst)
        right_queue.append("#")
        while len(right_queue) > 1:
            center = right_queue.popleft()
            corr_set.add(center)
            if center not in context_dict:
                context_dict[center] = set()
            left_context = " ".join(left_queue)
            right_context = " ".join(right_queue)
            context_dict[center].add((left_context,right_context))
            left_queue.append(center)
    for corr in corr_set:
        for phoneme in corr:
            phoneme_set.add(phoneme)
    if cfg.verbosity >= 20:
        print(sorted(list(corr_set)))
        for corr, lst in context_dict.items():
            print_rule(corr, lst, "=>")
    return

corr_containing_group_syms = set()

rule_lst = []

def print_rule(corr, context_set, rule_type):
    lst = ["   " + left + " _ " + right
           for left, right in context_set]
    if lst:
        ####print(corr, rule_type)
        strg = " ,\n".join(sorted(lst)).replace("#", ".#.") + " ;"
        rule_lst.append(corr + " " + rule_type + "\n" + strg)
        for left, right in context_set:
            sym_lst = left.split()
            sym_lst.extend(right.split())
            for sym in sym_lst:
                if sym not in corr_set:
                    corr_containing_group_syms.add(sym)
    #else:
    #    rule_lst.append(corr + " => _ ;")
    return

def identity_corr(strg):
    return strg.count(strg[0]) == len(strg)

def collect_corrs(strg, lst):
    global group_expansion_set
    #print(">>>", group_expansion_set, strg, lst) ###
    if lst:
        for phon in lst[0]:
            collect_corrs(strg + phon, lst[1:])
    else:
        if identity_corr(strg):
            strg = strg[0]
        if strg in corr_set:
            group_expansion_set.update({strg})
            #print("<<<", strg, group_expansion_set) ###
            return

def near_correspondence(corr1, corr2, zero):
    n = 0
    z = True
    for (c1, c2) in zip(corr1, corr2):
        if c1 != c2:
            n = n + 1
        elif c1 == zero:
            z = False
    return n == 1 and z

def near_correspondences(corr, zero):
    near_set = []
    corr_expanded = corr_width * corr if len(corr) == 1 else corr
    for c in corr_set:
        c_expanded = corr_width * c if len(c) == 1 else c
        if near_correspondence(c_expanded, corr_expanded, zero):
            near_set.append(c)
    return sorted(near_set)

def symbol_count(context_string):
    l = len(context_string.split())
    return l

def truncate_left(context_set, left_len):
    new_set = set()
    for left, right in context_set:
        left_lst = left.split()
        if len(left_lst) >= left_len:
            new_set.add((" ".join(left_lst[1:]), right))
        else:
            new_set.add((left, right))
    return new_set
            
def truncate_right(context_set, right_len):
    new_set = set()
    for left, right in context_set:
        right_lst = right.split()
        if len(right_lst) >= right_len:
            new_set.add((left, " ".join(right_lst[:right_len - 1])))
        else:
            new_set.add((left, right))
    return new_set

group_lst = []
group_sets = {}

def read_groups(fil):
    skipping = True
    for line_nl in fil:
        line = line_nl.strip()
        if skipping and line == "GROUPS":
            skipping = False
            continue
        elif skipping or not line or line.startswith("#"):
            continue
        elif line == "END":
            break
        line = line.split("#", maxsplit=1)[0].strip()
        lst = line.split()
        group_lst.append(("".join(lst[1:]), lst[0]))
        phon_set = set()
        for letter in lst[1:]:
            if letter in phoneme_set:
                phon_set.add(letter)
            elif letter in group_sets:
                phon_set.update(group_sets[letter])
            else:
                print("***", letter, "***", line_nl)
        group_sets[lst[0]] = phon_set
    if args.verbosity >= 10:
        print("group_lst:", group_lst)
    return

def reduce_string(input_string, symbol_string, set_symbol):
    global corr_width
    table = str.maketrans(symbol_string, len(symbol_string) * set_symbol)
    new_string = input_string.translate(table)
    new_string = new_string.replace(corr_width*set_symbol,set_symbol)
    return new_string

def reduce_context_set(context_set, symbol_string, set_symbol):
    new_set = set()
    for left, right in context_set:
        new_left = reduce_string(left, symbol_string, set_symbol)
        new_right = reduce_string(right, symbol_string, set_symbol)
        new_set.add((new_left, new_right))
    return new_set

def main():
    global group_expansion_set
    import argparse
    arpar = argparse.ArgumentParser("twol-histdiscov")
    arpar.add_argument(
        "algcognates",
        help="aligned cognates")
    arpar.add_argument(
        "-c", "--correspondence",
        help="one correspondence for which rules are proposed")
    arpar.add_argument(
        "-g", "--groups",
        help="file for alphabet and phoneme groups")
    arpar.add_argument(
        "-v", "--verbosity",
        help="level of  diagnostic output",
        type=int, default=0)
    args = arpar.parse_args()
    cfg.verbosity = args.verbosity

    read_aligned_cognates(args.algcognates)

    if args.groups:
        read_groups(open(args.groups, "r"))

    if args.correspondence:
        correspondences = [args.correspondence]
    else:
        correspondences = sorted(list(corr_set))

    for corr in correspondences:
        ###if len(corr) == 1:
        ###    continue
        positives = context_dict[corr]
        if cfg.verbosity >= 3:
            print_rule(corr, positives, "=> ! trivial")
        negatives = set()
        for near_corr in near_correspondences(corr, "Ø"):
            ctx_set = context_dict[near_corr]
            for x in ctx_set:
                negatives.add(x)
        negatives.difference(positives)
        if cfg.verbosity >= 4:
            print_rule(corr, negatives, "/<= ! trivial")

        # Truncate the right context
        while True:
            right_len = max([symbol_count(right)
                            for left, right in positives | negatives])
            if right_len <= 0:
                break
            new_positives = truncate_right(positives, right_len)
            new_negatives = truncate_right(negatives, right_len)
            if new_positives & new_negatives:
                break
            else:
                positives = new_positives
                negatives = new_negatives
                if cfg.verbosity >= 5:
                    print_rule(corr, positives, "=> ! right truncated")
                    print_rule(corr, negatives, "/<= ! right truncated")

        # Truncate the left context
        while True:
            left_len = max([symbol_count(left)
                            for left, right in positives | negatives])
            if left_len <= 0:
                break
            new_positives = truncate_left(positives, left_len)
            new_negatives = truncate_left(negatives, left_len)
            if new_positives & new_negatives:
                break
            else:
                positives = new_positives
                negatives = new_negatives
                if cfg.verbosity >= 5:
                    print_rule(corr, positives, "=> ! left truncated")
                    print_rule(corr, negatives, "/<= ! left truncated")

        # Reduce according to a list of phoneme groups
        succeeded_groups = set()
        all_group_set = set(group_sets.keys())
        for sym_str, set_sym in group_lst:
            gs = group_sets[set_sym]
            gsg = gs & all_group_set
            if (gsg) and (not gsg <= succeeded_groups):
                continue
            new_positives = reduce_context_set(positives, sym_str, set_sym)
            new_negatives = reduce_context_set(negatives, sym_str, set_sym)
            if not (new_positives & new_negatives):
                positives = new_positives
                negatives = new_negatives
                succeeded_groups.add(set_sym)
                if cfg.verbosity >= 5:
                    print_rule(corr, positives, "=> ! " + set_sym + " reduced")
                    print_rule(corr, negatives, "/<= ! " + set_sym + " reduced")

        if len(positives) <= len (negatives) or len(positives) == 0:
            print_rule(corr, positives, "=>")
            if cfg.verbosity > 3:
                print_rule(corr, negatives, "/<=")
        else:
            print_rule(corr, negatives, "/<=")
            if cfg.verbosity > 3:
                print_rule(corr, positives, "=>")
        
    #print("corr_containing_group_syms:", corr_containing_group_syms) ###
    #print("group_sets:", group_sets) ###
    for corr in sorted(list(corr_containing_group_syms)):
        if corr == "#":
            continue
        if len(corr) == 1:
            expanded_corr = corr_width * corr
        else:
            expanded_corr = corr
        lst = []
        for letter in expanded_corr:
            if letter in phoneme_set:
                lst.append(set(letter))
            elif letter in group_sets:
                lst.append(group_sets[letter])
        #print("lst", corr, lst) ###
        group_expansion_set = set()
        collect_corrs("", lst)
        #print("group_expansion_set:", group_expansion_set) ###
        ls = sorted(list(group_expansion_set))
        print(corr, "=", " | ".join(ls), ";")

    for rule in rule_lst:
        print(rule)
    return

if __name__ == "__main__":
    main()
